'''
© Copyright ArdillaByte Inc. 2023

-----------------------------------------------
Class to XXXX.
-----------------------------------------------
'''
import numpy as np
from typing import List, Dict, Optional, Tuple, Union
from copy import deepcopy
import torch
import torch.nn.functional as F

from chaparral.modelos.constantes import NOMBRES_CLUSTERS


def a_tabla(cadena):
    c = '<table>\n<tbody>\n'
    c += cadena
    c += '</tbody>\n</table>'
    return c

class Quimico:
	'''
	Objeto quimico

	Atributos:
		- nombre
		- unidades de medida
	'''
	def __init__(
				self, 
				nombre:str, 
				unidades:str
			) -> None:
		self.nombre = nombre
		self.unidades = unidades

	def __str__(self) -> str:
		return f'{self.nombre} (unidades:{self.unidades})'

	def in_carga(
				self, 
				carga_quimicos:List[any]
			) -> Union[None, int]:
		'''Determina si self está en una lista y en caso positivo devuelve el índice en la lista'''
		nombres_quimicos = [quimico.nombre for quimico in carga_quimicos]
		if self.nombre in nombres_quimicos:
			return nombres_quimicos.index(self.nombre)
		else:
			return None


class QuimicoEn(Quimico):

	def __init__(
				self, 
				nombre: str, 
				unidades: str,
				nivel_actual: float,
				velocidad_consumo: float
			) -> None:
		super().__init__(nombre, unidades)
		self.nivel_actual = nivel_actual
		self.nivel_inicial = nivel_actual
		self.velocidad_consumo = velocidad_consumo

	def __str__(self) -> str:
		''' Genera una cadena en formato html con el nombre del químico, su nivel y velocidad'''
		cadena = f'<td> {self.nombre} </td> '
		cadena += f'<td> {round(self.nivel_actual, 2)} </td> '
		if self.velocidad_consumo is not None:
			cadena += f'<td> {round(self.velocidad_consumo, 2)} </td>'
		return cadena

	def actualizar(
				self, 
				unidades_temporales:float
			) -> None:
		self.nivel_actual -= self.velocidad_consumo * unidades_temporales
		self.nivel_actual = max(0, self.nivel_actual)
	
	def a_tensor(self) -> torch.tensor:
		if self.velocidad_consumo is not None:
			return torch.tensor([self.nivel_actual, self.velocidad_consumo], dtype=torch.float32)
		else:
			return torch.tensor([self.nivel_actual], dtype=torch.float32)			


class Pozo:
	'''
	Objeto pozo

	Atributos:
		- nombre
		- si está activo o no
		- lista de químicos en pozo
	'''
	def __init__(
				self,
				nombre:str,
				quimicos:List[QuimicoEn]
			) -> None:
		self.nombre = nombre
		self.quimicos = quimicos	
		self.num_quimicos = len(quimicos)

	def __str__(self) -> str:
		cadena = f'\n<tr>\n\t<td> {self.nombre} </td> '
		inicial = True
		for quimico in self.quimicos:
			if inicial:
				cadena += str(quimico) + '\n</tr> '	
				inicial = False
			else:
				cadena += ' \n<tr>\n\t<td>&nbsp;</td> ' + str(quimico) + '\n</tr>'	
		return cadena

	def actualizar(
				self,
				unidades_temporales:float
			) -> None:
		for quimico in self.quimicos:
			quimico.actualizar(unidades_temporales)

	def a_tensor(self) -> torch.tensor:
		tensores = [quimico.a_tensor() for quimico in self.quimicos]
		return torch.cat(tensores)


class Cluster:
	'''
	Objeto cluster

	Atributos:
		- nombre
		- lista de pozos
		- coordenadas del cluster
		# - lista de clusters vecinos
	'''
	def __init__(
				self, 
				nombre:str, 
				pozos:List[str],
				coordenadas:Tuple, 
				# vecinos:List[str]
			) -> None:
		self.nombre = nombre
		# -----------------------------------
		# Los pozos del cluster
		# -----------------------------------
		self.pozos = pozos
		self.num_pozos = len(self.pozos)
		# -----------------------------------
		# Atributos geográficos
		# -----------------------------------
		self.coordenadas = coordenadas # Coordenadas geográficas del cluster
		# self.vecinos = vecinos # Rutas que conectan el cluster con otros clusters/bodega
		
	def actualizar(
			self, 
			unidades_temporales:int
			) -> None:
		'''
		Actualiza el cluster simulando que transcurren las unidades temporales dadas

		Input:
			- unidades_temporales (int), la cantidad de unidades de tiempo que transcurren.
		'''
		for pozo in self.pozos:
			pozo.actualizar(unidades_temporales)
	
	def __str__(self) -> str:
		'''Genera una cadena en formato html con el nombre de cada químico y su cantidad actual'''
		cadena = '<tr>\n\t<td>Cluster</td> \n</tr>'
		cadena += f'\n<tr>\n\t<td> {self.nombre} </td> '
		inicial = True
		for pozo in self.pozos:
			if inicial:
				str_pozo = '<tr>\n\t<td>Pozo</td> <td>Químico</td> <td>Cantidad</td> <td>Vel.Cons</td> \n</tr>' + str(pozo)
				cadena += ' <td> ' + a_tabla(str_pozo) + ' </td> \n</tr> '	
				inicial = False
			else:
				str_pozo = '<tr>\n\t<td>Pozo</td> <td>Químico</td> <td>Cantidad</td> <td>Vel.Cons</td> \n</tr>' + str(pozo)
				cadena += ' <tr>\n\t<td>&nbsp;</td> ' + ' <td> ' + a_tabla(str_pozo) + ' </td> \n</tr> '	
		return a_tabla(cadena)
	
	def cluster_a_tensor(self) -> torch.tensor:
		'''Genera un tensor plano donde, por cada químico, incluye 
			el nivel actual,
			la velocidad de consumo.
		'''
		tensores = [pozo.a_tensor() for pozo in self.pozos]
		assert(len(tensores) > 0), f'¡Error!: ¡Cluster {self.nombre} no tiene pozos!'
		return torch.cat(tensores)


class Ruta:
	'''
	Objeto que representa la ruta que está siguiendo 
	un camión para ir de un cluster a otro.

	Atributos:
		- nombre del cluster inicial
		- nombre del cluster objetivo
		- longitud del camino en kilómetros
		- velocidad promedio de tránsito de un camión en kilómetros por unidad temporal
	'''
	def __init__(
				self, 
				inicial:Cluster, 
				objetivo:Cluster,
				longitud:float,
				velocidad_promedio:float,
			) -> None:
		self.inicial = inicial
		self.objetivo = objetivo
		self.nombre = f'{self.inicial.nombre} <-> {self.objetivo.nombre}'
		self.longitud = longitud
		self.velocidad_promedio = velocidad_promedio
		self.porcentaje_recorrido = 0

	def __str__(self) -> str:
		cadena = f'Ruta {self.nombre} por {round(self.longitud,2)}km a {self.velocidad_promedio}km/hora'
		cadena += '\n\t' + f'El camion ha recorrido el {round(self.porcentaje_recorrido, 2)}%'
		return cadena

	def actualizar(self, unidades_temporales:float) -> None:
		'''
		Mueve el camión por la ruta de acuerdo a la
		velocidad promedio de la ruta por la cantidad
		de unidades_temporales dada 
		'''
		# Encuentra la distancia recorrida después de unidades temporales
		distancia_recorrida = self.velocidad_promedio * unidades_temporales
		longitud_faltante =  self.longitud * (100 - self.porcentaje_recorrido) / 100
		if  distancia_recorrida >= longitud_faltante:
			# El camión llega al final de la ruta
			self.porcentaje_recorrido = 100
		else:
			self.porcentaje_recorrido = (1 - (longitud_faltante - distancia_recorrida) / self.longitud) * 100

	def fin_recorrido(self) -> bool:
		# Determina si el camino fue recorrido
		return self.porcentaje_recorrido == 100
	
	def obtener_unidades_temporales_faltantes(self) -> float:
		'''
		Determina cuántas unidades temporales faltan para terminar la ruta
		de acuerdo a la velocidad promedio del camino y la longitud faltante
		'''
		longitud_faltante =  self.longitud * (100 - self.porcentaje_recorrido) / 100
		return longitud_faltante / self.velocidad_promedio


class PseudoRutaCamionEstacionario(Ruta):
	'''
	Clase que simula una ruta pero establece
	los métodos de ruta para cuando el camión
	está estacionario.
	'''
	def __init__(
				self, 
				inicial: Cluster, 
				objetivo: Cluster, 
				longitud: float, 
				velocidad_promedio: 
				float
			) -> None:
		super().__init__(inicial, objetivo, longitud, velocidad_promedio)

	def __str__(self) -> str:
		return f'Camión estacionado en {self.inicial.nombre}'
	
	def actualizar(self, unidades_temporales: float) -> None:
		return None
	
	def fin_recorrido(self) -> bool:
		return True
	
	def obtener_unidades_temporales_faltantes(self) -> float:
		return np.Inf
	

class Camion:
	'''
	Objeto camion
	
	Atributos:
		- nombre
		- capacidad de carga
		- lista de químicos que carga
		- ubicación
	'''
	def __init__(
				self, 
				nombre:str, 
				capacidad:float, 
				carga_quimicos:List[QuimicoEn],
				ubicacion:int,
				ruta:Union[None, Ruta]
			) -> None:
		self.nombre = nombre
		self.capacidad = capacidad
		self.carga_quimicos = carga_quimicos
		self.num_quimicos = len(self.carga_quimicos)
		self.ubicacion = ubicacion
		self.ruta = None

	def __str__(self) -> str:
		cadena = '<tr>\n\t<td>Camion</td> <td>Ubicado en</td> <td>Químico</td> <td>Carga</td>\n</tr>'
		cadena += f'\n<tr>\n\t<td> {self.nombre} </td> <td>{NOMBRES_CLUSTERS[self.ubicacion]}</td> '
		inicial = True
		for quimico in self.carga_quimicos:
			if inicial:
				cadena += str(quimico) + '\n</tr> '	
				inicial = False
			else:
				cadena += ' \n<tr>\n\t<td>&nbsp;</td> <td>&nbsp;</td> ' + str(quimico) + '\n</tr>'	
		return cadena

	def camion_a_tensor(
				self, 
				num_clusters:Optional[int]
			) -> torch.tensor:
		tensores = [quimico.a_tensor() for quimico in self.carga_quimicos]
		tensor_ubicacion = torch.tensor(self.ubicacion)
		tensores += [F.one_hot(tensor_ubicacion, num_classes=num_clusters)]
		return torch.cat(tensores)


'''
© Copyright ArdillaByte Inc. 2023
-----------------------------------------------
Implements:
    - the base tabular agent, Agent
    - the base agent for Q approximation, AgentCS
    - the base agent for Q approximations based on Neural Networks, AgentNN
    - the base agent for policy learning, PolicyAgent
-----------------------------------------------
'''

import numpy as np
from typing import Dict
import json
from copy import deepcopy
from typing import Optional
from pathlib import Path
from termcolor import colored
from collections import deque


class AgentNN() :
    '''
    Super class of agents with Q approximation
    using Neural Networks.
    '''

    def __init__(self, parameters:dict):
        self.parameters = parameters
        self.nA = self.parameters['nA']
        self.gamma = self.parameters['gamma']
        self.epsilon = self.parameters['epsilon']
        self.NN = self.parameters['NN']
        assert(hasattr(self.NN, 'predict')), 'NN must be an object with a predict() method'
        assert(hasattr(self.NN, 'values_vector')), 'NN must be an object with a values_vector() method'
        assert(hasattr(self.NN, 'learn')), 'NN must be an object with a learn() method'
        assert(hasattr(self.NN, 'save')), 'NN must be an object with a save() method'
        assert(hasattr(self.NN, 'load')), 'NN must be an object with a load() method'
        assert(hasattr(self.NN, 'reset')), 'NN must be an object with a reset() method'
        self.max_len = self.parameters['max_len']
        self.states = deque(maxlen=self.max_len)
        self.actions = deque(maxlen=self.max_len)
        self.rewards = deque([np.nan], maxlen=self.max_len)
        self.dones = deque([np.nan], maxlen=self.max_len)
        self.seed = None
        self.debug = False
        # Create model file
        self.model_folder = Path.cwd() / Path('..').resolve() / Path('..').resolve() / Path('models', 'MLP')
        self.model_folder.mkdir(parents=True, exist_ok=True)
        self.model_file = self.model_folder / Path('mlp.pt')

    def make_decision(self, state:Optional[any]=None):
        '''
        Agent makes an epsilon greedy accion based on Q values.
        '''
        if np.random.uniform(0,1) < self.epsilon:
            return np.random.choice(range(self.nA))
        else:
            if state is None:
                state = self.states[-1]
            return self.argmaxQ(state)        

    def argmaxQ(self, state):
        '''
        Determines the action with max Q value in state s
        '''
        Qvals = self.NN.values_vector(state)
        maxQ = max(Qvals)
        opt_acts = [a for a, q in enumerate(Qvals) if np.isclose(q, maxQ)]
        if self.seed is not None:
            np.random.seed(self.seed)
        try:
            argmax = np.random.choice(opt_acts)
            return argmax
        except Exception as e:
            print('')
            print(colored('%'*50, 'red'))
            print(colored(f'Error in argmaxQ ====>', 'red'))
            print(colored(f'state:\n\t{state}', 'red'))
            print('')
            print(colored(f'Qvals:{Qvals}', 'red'))
            print(colored(f'len:{len(Qvals)} --- type:{type(Qvals)}', 'red'))
            print('')
            print(colored(f'maxQ:{maxQ}', 'red'))
            print(colored(f'type:{type(maxQ)}', 'red'))
            print('')
            print(colored(f'opt_acts:{opt_acts}', 'red'))
            print(colored(f'opt_acts:{[a for a, q in enumerate(Qvals) if np.isclose(q, maxQ)]}', 'red'))
            print(colored('%'*50, 'red'))
            print('')
            raise Exception(e)            

    def update(self, next_state, reward, done) -> None:
        '''
        Agent updates its NN according to a model.
            TO BE OVERWRITTEN BY SUBCLASS  
        '''
        pass

    def restart(self) -> None:
        '''
        Restarts the agent for a new trial.
        '''
        self.states = deque(maxlen=self.max_len)
        self.actions = deque(maxlen=self.max_len)
        self.rewards = deque([np.nan], maxlen=self.max_len)
        self.dones = deque([np.nan], maxlen=self.max_len)
        self.NN.restart()

    def reset(self) -> None:
        '''
        Resets the agent for a new simulation.
        '''
        self.restart()
        # Resets the NN
        self.NN.reset()
    
    def save(self, file:Path) -> None:
        self.NN.save(file=file)

    def load(self, file:Path) -> None:
        self.NN.load(file=file)



'''
© Copyright ArdillaByte Inc. 2023

-----------------------------------------------
Clase para correr el campo de pozos con sus
clusters y camiones.
-----------------------------------------------
'''
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import networkx as nx
import pandas as pd
import numpy as np
from typing import List, Dict, Optional, Tuple, Union
from IPython.display import HTML
from copy import deepcopy
import torch
from prettytable import PrettyTable
import pickle
from pathlib import Path
from collections import defaultdict

from chaparral.modelos.clases import (
	Cluster, 
	Camion, 
	Ruta, 
	PseudoRutaCamionEstacionario
)
from chaparral.modelos.constantes import (
	FACTOR_RECOMPENSA,
	RECOMPENSA_MOVIMIENTO_NO_PERMITIDO,
	HORAS_HABILES,
	PENALIZACION_PASADAS_HORAS_HABILES,
	RUIDO,
	VALOR_CRITICO_TOLERANCIA,
	VALOR_CRITICO_CARGA,
	MEDIA_X,
	DESVEST_X,
	MEDIA_Y,
	DESVEST_Y,
	FACTOR_RENDER_X,
	FACTOR_RENDER_Y
)


class CampoPozos:
	'''
	El entorno de los clusters unidos por caminos.
 
	Atributos:
		- Lista de clusters
		- Lista de caminos
		- Lista de camiones
		- Datos de las rutas
	'''
	def __init__(
				self, 
				clusters:List[Cluster], 
				datos_rutas:Dict[str, Dict[str,float]],
				camiones:List[Camion]
			) -> None:
		# -------------------------------
		# Datos provenientes de clusters
		# -------------------------------
		self.clusters = clusters
		self.backup_clusters = deepcopy(clusters)
		self.num_clusters = len(self.clusters)
		self.punto_partida = clusters[0].nombre
		self.dict_clusters = {cluster.nombre:p for p, cluster in enumerate(self.clusters)}
		self.coordenadas = {cluster.nombre:cluster.coordenadas for cluster in self.clusters}
		# -------------------------------
		# Datos provenientes de rutas
		# -------------------------------
		self.datos_rutas = datos_rutas
		# -------------------------------
		# Datos referentes al turno
		# -------------------------------
		self.tiempo_transcurrido = 0
		# -------------------------------
		# Datos provenientes de camiones
		# -------------------------------
		self.camiones = camiones
		self.actualizar_camiones()
		self.backup_camiones = deepcopy(camiones)
		self.num_camiones = len(self.camiones)
		# -------------------------------
		# Para usar en entrenamiento
		# -------------------------------
		self.usar_ruido = False
		# -------------------------------
		# Inicializa el estado
		# -------------------------------
		self.estado = self.reset()
		self.nS = self.estado.shape[0]
		self.comienzo = True # El entorno recién se inicializa
		# -------------------------------
		# Crea el bookkeeping de las acciones
		# -------------------------------
		# Las acciones posibles es decirle a cada camión
		# a qué cluster ir
		self.nA = self.num_clusters * self.num_camiones
		# -------------------------------
		# Para usar en debugging
		# -------------------------------
		self.debug = False
		# -------------------------------
		# Para usar en render
		# -------------------------------
		self.render_mode = 'rgb_array'

	def reset(self):
		'''
		Reinicia el entorno y devuelve el estado inicial
		'''
		self.clusters = deepcopy(self.backup_clusters)
		self.camiones = deepcopy(self.backup_camiones)
		estado = self.obtener_estado()
		self.comienzo = True # El entorno recién se inicializa
		self.tiempo_transcurrido = 0
		if self.usar_ruido:
			self.incluir_ruido()
		return estado
	
	def step(self, accion:int) -> Tuple[torch.tensor, float, bool]:
		'''
		Avanza el entorno de acuerdo a la tupla de acciones.

		Input:
			- accion, índice de la lista de qué ruta toma cada camión

		Output:
			- tripla con nuevo estado, recompensa y done
		'''
		# El entorno comenzó
		self.comienzo = False
		# Decodifica la acción como índice de lista de movimientos
		movimientos_camiones = self.accion_a_movimientos(accion)
		# Verifica que todos los movimientos sean válidos
		verificacion_movimientos = self._check_movimientos(movimientos_camiones)
		if not verificacion_movimientos:
			# Alguno de los movimientos no es permitido.
			# El entorno debe seguir en el mismo estado 
			# y devolver castigo mediante la tupla: 
			# (mismo estado, recompensa negativa, entorno sigue activo)
			return self.estado, RECOMPENSA_MOVIMIENTO_NO_PERMITIDO, True
		# Si un camión debe comenzar una nueva ruta, inicializar en la lista atributo
		self._inicializar_rutas(movimientos_camiones)
		# Obtenemos cuántas unidades temporales se demora cada
		# camión en su respectiva ruta
		unidades = self._obtener_unidades_temporales()
		# Las unidades temporales a transcurrir son
		# el menor tiempo para terminar una ruta
		if self.debug:
			print('Unidades temporales de las rutas:', unidades)
		unidades_temporales = min(unidades)
		if unidades_temporales == np.inf:
			# Si el mínimo es inf, significa que ningún camión se mueve
			# Esto solo debe ocurrir cuando el entorno termina
			return self.estado, RECOMPENSA_MOVIMIENTO_NO_PERMITIDO, True
		# para bookkeeping de espacio de estados
		self._unidades_temporales = unidades_temporales
		if self.debug:
			print(f'Unidades temporales que transcurren: {unidades_temporales}')
		# Actualiza el estado
		self._actualizar_estado(unidades_temporales)
		nuevo_estado = self.obtener_estado()
		# Encuentra la recompensa y el done
		recompensa, done = self._encontrar_recompensa_y_done(unidades_temporales)
		return nuevo_estado, recompensa, done
	
	def obtener_estado(
				self,
				noise_:Optional[float]=RUIDO
			) -> torch.tensor:
		'''
		Obtiene el estado actual a partir de los atributos del entorno
		'''
		tensor_tiempo = [torch.tensor([self.tiempo_transcurrido], dtype=torch.float32)]
		tensores_clusters = [cluster.cluster_a_tensor() for cluster in self.clusters[1:]]
		tensores_camiones = [camion.camion_a_tensor(self.num_clusters) for camion in self.camiones]
		tensores = tensor_tiempo + tensores_clusters + tensores_camiones
		tensores = torch.cat(tensores)
		# Incluimos algo de ruido en el tensor
		shape = tensores.shape
		RUIDO = torch.rand(shape) * noise_
		return torch.add(tensores, RUIDO)

	def _actualizar_estado(
				self, 
				unidades_temporales:float
			) -> None:
		'''
		Actualiza el estado con la lógica siguente:
			1. Actualiza los clusters
			2. Actualiza los camiones en su respectiva ruta
			3. Si un camión llega al final de la ruta, descarga lo que pueda
		'''
		# 1. Pedimos a cada cluster que actualize con la cantidad de unidades temporales
		for cluster in self.clusters:
			cluster.actualizar(unidades_temporales)
		# 2. Pedimos a cada ruta que actualize con la cantidad de unidades temporales
		for camion in self.camiones:
			camion.ruta.actualizar(unidades_temporales)
			# 3. Verificar si el camión que va en la ruta llegó al final
			if camion.ruta.fin_recorrido():
				nombre_cluster_llegada = camion.ruta.objetivo.nombre
				if self.debug:
					print(f'Camión {camion.nombre} llegó a cluster {nombre_cluster_llegada}')
				# Descarga químicos en cluster
				indice_cluster = self.dict_clusters[nombre_cluster_llegada]
				self._descargar_camion_en_cluster(
					nombre_cluster=nombre_cluster_llegada, 
					camion=camion
				)
				# Actualizar la posición del camión
				camion.ubicacion = self.dict_clusters[nombre_cluster_llegada]
				# Dejar en espera de nueva ruta para el camión
				camion.ruta = None
		# para bookkeeping del turno
		self.tiempo_transcurrido += unidades_temporales

	def _encontrar_recompensa_y_done(self, unidades_temporales) -> Tuple[float, bool]:
		'''Encuentra la recompensa en el estado obtenido'''
		# # Chequeamos si la cantidad de algún químico en algún cluster es negativa
		# for cluster in self.clusters[1:]:
		# 	for pozo in cluster.pozos:
		# 		for quimico in pozo.quimicos:
		# 			if quimico.nivel_actual < quimico.velocidad_consumo * VALOR_CRITICO_TOLERANCIA:
		# 				if self.debug:
		# 					print('¡Se infringieron los porcentajes críticos!')
		# 					print(f'\tCluster {cluster.nombre}--{pozo.nombre}--{quimico.nombre} con niveles: {round(quimico.nivel_actual, 2)} (mínimo: {round(quimico.velocidad_consumo * VALOR_CRITICO_TOLERANCIA, 2)})')
		# 				# Por lo menos un químico está por debajo del nivel crítico => penalizar al agente
		# 				return -1000, True
		# Chequeamos si todos los camiones están en bodega
		# y no es el estado incial
		# Si el camión tomó más de las horas hábiles, penalizar
		penalizacion = 0
		if self.tiempo_transcurrido > HORAS_HABILES:
			penalizacion = PENALIZACION_PASADAS_HORAS_HABILES
		if not self.comienzo and np.all([camion.ubicacion == 0 for camion in self.camiones]):
			# Entorno terminado
			# # Encontramos la recompensa como la suma de las diferencias 
			# # entre los niveles actuales y los requeridos (todos estos valores
			# # deben ser negativos)
			# recompensas = [np.subtract(cluster.niveles_actuales, 1) for cluster in self.clusters if cluster.activo]
			# Encontramos la recompensa como la suma de los niveles por encima del 1%
			recompensas = list()
			for cluster in self.clusters:
				for pozo in cluster.pozos:
					for quimico in pozo.quimicos:
						if quimico.velocidad_consumo > 0:
							recompensas.append(quimico.nivel_actual // quimico.velocidad_consumo)
			if self.debug:
				print(f'Las diferencias de químicos son: {recompensas}')
			recompensa = sum(recompensas) * FACTOR_RECOMPENSA
			recompensa += -1 * unidades_temporales + penalizacion
			return recompensa, True
		return -1 * unidades_temporales + penalizacion, False	

	def _obtener_unidades_temporales(self) -> Union[None, List[int]]:
		'''
		Obtiene la cantidad de unidades temporales requeridas 
		para terminar cada ruta de acuerdo a los movimientos
		de los camiones dado en el argumento

		Input:
			- movimientos_camiones, la lista de qué ruta toma cada camión

		Output:
			- Lista de unidades temporales o None en caso de acción inválida
		'''
		unidades = list()
		for camion in self.camiones:
			unidad = camion.ruta.obtener_unidades_temporales_faltantes()
			unidades.append(unidad)
		return unidades

	def _inicializar_rutas(self, movimientos_camiones:List[int]) -> None:
		'''Verifica si camión está en la ruta o si no inicializa'''
		for indice_camion, movimiento in enumerate(movimientos_camiones):
			ruta = self.camiones[indice_camion].ruta
			if ruta is None:
				self._iniciar_ruta(indice_camion, movimiento)

	def _iniciar_ruta(self, indice_camion, movimiento) -> None:
		'''Obtiene la ruta que debe seguir el camión y modifica la lista atributo'''
		# Obtiene dónde está el camión
		indice_inicial = self.camiones[indice_camion].ubicacion
		inicial = self.clusters[indice_inicial]
		# Obtiene para dónde va el camión de acuerdo al movimiento
		objetivo = self.clusters[movimiento]
		# Verifica si el movimiento estipula que el camión debe quedarse donde está
		if inicial.nombre != objetivo.nombre:
			# Crea la ruta
			ruta = Ruta(
				inicial=inicial,
				objetivo=objetivo,
				tiempo=self.datos_rutas[inicial.nombre][objetivo.nombre]
			)
		else:
			# Debemos crear una pseudoruta porque el movimiento
			# determina que el camión se queda estacionario
			ruta = PseudoRutaCamionEstacionario(
				inicial=inicial,
				objetivo=objetivo,
				tiempo=np.Inf
			)
		self.camiones[indice_camion].ruta = ruta

	def _check_movimientos(self, movimientos_camiones:List[int]) -> bool:
		'''
		Verifica que todos los movimientos sean válidos
		'''
		for indice_camion, movimiento in enumerate(movimientos_camiones):
			verificacion = self._check_accion(
				indice_camion=indice_camion,
				movimiento=movimiento
			)
			if not verificacion:
				return False
		return True

	def _check_accion(self, indice_camion:int, movimiento:int) -> bool:
		'''
		Verifica que, si el camión está en ruta, el movimiento
		coincide con la dirección de la ruta. También es posible
		que el camión se quede estacionario.
		'''
		# Obtiene el nombre del culster dado por el movimiento
		a_cluster = self.clusters[movimiento].nombre
		# Si camión en transito, accion debe coincidir con destino
		ruta = self.camiones[indice_camion].ruta
		if ruta is None:
			# Camión no está en tránsito
			check = True
		else:
			# Camión en tránsito
			destino = ruta.objetivo.nombre
			check = destino == a_cluster
		if self.debug:
			print(f'Movimiento en ruta: {check}')
		return check
	
	def _descargar_camion_en_cluster(
				self, 
				nombre_cluster:str,
				camion:Camion
			) -> None:
		'''Descarga los químicos en el cluster'''
		# Carga el objeto cluster
		indice_cluster = self.dict_clusters[nombre_cluster]
		cluster = self.clusters[indice_cluster]
		# Itera sobre los pozos del cluster
		for pozo in cluster.pozos:
			# Itera sobre los químicos del pozo
			for quimico in pozo.quimicos:
				# Determina si el quimico está activo (velocidad de consumo mayor a 0)
				if quimico.velocidad_consumo > 0:
					# Determina si el camion tiene el químico
					indice_quimico = quimico.in_carga(camion.carga_quimicos)
					if indice_quimico is not None:
						# Encuentra la cantidad que falta del químico
						# como la diferencia entre el valor requerido y el actual
						falta = max(quimico.velocidad_consumo * VALOR_CRITICO_CARGA - quimico.nivel_actual, 0)
						carga = camion.carga_quimicos[indice_quimico].nivel_actual
						# Calcula la cantidad a descargar del camión
						if 0 <= falta <= carga:
							descarga = falta
						elif falta > carga:
							descarga = carga
						else:
							raise Exception(f'¡Valor de descarga inválido!\n\t carga:{carga} --- falta:{falta}')
						if self.debug:
							print(f'El camión {camion.nombre} descarga {round(descarga, 2)} de {quimico.nombre} (requerimiento: {round(falta, 2)})')
						# Actualiza los niveles de cluster y camión
						quimico.nivel_actual += descarga
						camion.carga_quimicos[indice_quimico].nivel_actual -= descarga
	
	def accion_a_movimientos(self, accion:int) -> List[int]:
		"""Toma una acción y devuelve la lista correspondiente"""
		shape = tuple([self.num_clusters] * self.num_camiones)
		return list(np.unravel_index(accion, shape))
	
	def movimientos_a_accion(self, movimientos:Tuple[int]) -> int:
		"""Toma una lista de movimientos y devuelve la acción como índice"""
		shape = tuple([self.num_clusters] * self.num_camiones)
		return np.ravel_multi_index(movimientos, shape)
	
	def incluir_ruido(self):
		"""Modifica los valores de quimicos para facilitar aprendizaje"""
		for cluster in self.clusters:
			for pozo in cluster.pozos:
				for quimico in pozo.quimicos:
					quimico.nivel_actual = np.random.normal(loc=100, scale=10)
					quimico.velocidad_consumo = max(0.001, np.random.normal(loc=1e-1, scale=1e-2))
		for camion in self.camiones:
			for quimico in camion.carga_quimicos:
				quimico.nivel_actual = np.random.normal(loc=220, scale=10)
		
	def generar_rutas_de_politica(
				self, 
				agente:any, 
				max_steps:Optional[int]=10
			) -> List[Tuple[Dict,PrettyTable,PrettyTable]]:
		"""
		Genera las rutas usando un agente
		
		Input:
			- politica, una función que genera movimientos de los camiones en un estado dado
		
		Output:
			- dict_rutas, un diccionario con la ruta de cada camión
		"""
		# Hace un backup del valor de epsilon
		backup_epsilon = deepcopy(agente.epsilon)
		# Apaga la exploración del agente
		agente.epsilon = 0
		# Inicia los registros
		num_steps = 0
		done = False
		estado = self.reset()
		accion = agente.make_decision(estado)
		movimientos_camiones = self.accion_a_movimientos(accion)
		dict_rutas = {camion.nombre:[movimientos_camiones[indice_camion]] for indice_camion, camion in enumerate(self.camiones)}
		# Usa el agente óptimo para recorrer el campo de pozos
		while not done and num_steps < max_steps:
			num_steps += 1
			estado, recompensa, done = self.step(accion)
			accion = agente.make_decision(estado)
			movimientos_camiones = self.accion_a_movimientos(accion)
			for indice_camion, camion in enumerate(self.camiones):
				dict_rutas[camion.nombre].append(movimientos_camiones[indice_camion])
		table = PrettyTable(['Camión', 'Cluster'])
		table_quimicos = PrettyTable(['Camión', 'Quimico', 'Cantidad'])
		for camion in self.camiones:
			# Crea la tabla de químicos a llevar
			for quimico in camion.carga_quimicos:
				diferencia = quimico.nivel_inicial - quimico.nivel_actual
				if diferencia > 0:
					table_quimicos.add_row([camion.nombre, quimico.nombre, diferencia])
			# Crea la tabla de rutas
			table.add_row([camion.nombre, self.punto_partida])
			for i in range(num_steps):
				indice_cluster = dict_rutas[camion.nombre][i]
				a_cluster = self.clusters[indice_cluster].nombre
				table.add_row([camion.nombre, a_cluster])
		# Reinstaura el valor de epsilon
		agente.epsilon = backup_epsilon
		return (dict_rutas, table, table_quimicos)
	
	def render(self):
		G = nx.Graph()
		G.add_nodes_from(self.coordenadas.keys())
		order = {cluster:self._transformar_coordenadas(self.coordenadas[cluster]) for cluster in self.coordenadas.keys()}
		options = {
			'node_color': 'green',
			'node_size': 10,
			'verticalalignment':'bottom',
			'font_size': 12
		}
		fig = plt.figure(figsize=(15,15))
		ax = fig.add_subplot()
		nx.draw_networkx(G, pos=order, font_weight='bold', **options)
		for c, camion in enumerate(self.camiones):
			camion_en_pozo = camion.ubicacion
			coordenadas = self._transformar_coordenadas(self.clusters[camion_en_pozo].coordenadas)
			circle1 = patches.Circle(coordenadas, radius=0.25, color='green')
			ax.add_patch(circle1)
		ax.axis('off')
		fig.canvas.draw()
		string = fig.canvas.renderer.buffer_rgba()
		array = np.array(string)
		if self.render_mode == 'rgb_array':
			return array
		else:
			plt.show()

	def _transformar_coordenadas(
				self,
				coordenadas:List[float]
			) -> List[float]:
		x = (coordenadas[0] - MEDIA_X) / DESVEST_X * FACTOR_RENDER_X
		y = (coordenadas[1] - MEDIA_Y) / DESVEST_Y * FACTOR_RENDER_Y
		return x, y

	def __str__(self) -> str:
		cadena = '<table> <tbody>'
		for cluster in self.clusters:
			cadena += str(cluster)
		cadena += '</tbody> </table>'
		cadena += '<tr> <td>&nbsp;</td> <td>&nbsp;</td> <td>&nbsp;</td> </tr>'
		cadena += '<table> <tbody>'
		for camion in self.camiones:
			cadena += str(camion)
		cadena += '</tbody> </table>'
		return cadena

	def state_visualizer(self, state):
		"""Retorna una representación visual del estado"""
		# html_table = str(self)
		# table_MN = pd.read_html(html_table)
		# return table_MN
		return str(state.tolist())

	def action_visualizer(self, action):
		"""Retorna una representación visual de la acción"""
		movimientos_camiones = self.accion_a_movimientos(action)
		lista_movimientos = [self.clusters[indice_cluster].nombre for indice_cluster in movimientos_camiones]
		return lista_movimientos
	
	def save(
				self,
				archivo:Optional[Union[None, str]]=None
			) -> None:
		directorio = Path.cwd() / Path('..').resolve() / Path('data', 'entornos')
		directorio.mkdir(parents=False, exist_ok=True)
		if archivo is not None:
			path_archivo = directorio / Path(archivo)
		else:
			path_archivo = directorio / Path('entorno.pkl')
		with open(path_archivo, 'wb') as file:
			pickle.dump(self, file)
			file.close()

	def actualizar_camiones(self) -> None:
		"""Actualiza la carga de los camiones"""
		# Actualiza las estimaciones de las velocidades de
		# consumo de los químicos en el campo
		estimaciones = self.estimar_velocidades_consumo_clusters()
		nombres_quimicos = list(estimaciones.keys())
		for cluster in self.clusters:
			for pozo in cluster.pozos:
				for quimico in pozo.quimicos:
					nombres_quimicos.append(quimico.nombre)
		# Actualiza camiones con estimaciones y valor crítico de carga
		for camion in self.camiones:
			camion.ubicacion = 0
			camion.ruta = None
			camion.nombres_clusters = [cluster.nombre for cluster in self.clusters]
			camion.carga_quimicos = [quimico for quimico in camion.carga_quimicos if quimico.nombre in nombres_quimicos]
			for quimico in camion.carga_quimicos:
				quimico.nivel_actual = estimaciones[quimico.nombre] * VALOR_CRITICO_CARGA
				quimico.nivel_inicial = quimico.nivel_actual

	def estimar_velocidades_consumo_clusters(self) -> Dict[str, float]:
		"""
		Estima las velocidades de consumo de los químicos
		en todos los pozos del campo.
		"""
		velocidades = defaultdict(list)
		estimaciones = dict()
		for cluster in self.clusters:
			for pozo in cluster.pozos:
				for quimico in pozo.quimicos:
					velocidades[quimico.nombre].append(quimico.velocidad_consumo)
		nombres_quimicos = list(velocidades.keys())
		for nombre_quimico in nombres_quimicos:
			estimaciones[nombre_quimico] = np.mean(velocidades[nombre_quimico])
		return estimaciones		


import cv2 as cv
import numpy as np

# read image
main_path = 'data/Mapa General Campo Chichimene.png'
pozos_path = 'data/clusters_prv.png'
via1_path = 'data/pavimentada_prv.png'
via2_path = 'data/pavimentada_angosta_prv.png'

img_main = cv.imread(main_path)
img_pozos = cv.imread(pozos_path)
img_via1 = cv.imread(via1_path)
img_via2 = cv.imread(via2_path)
# crop = img[1210:1225, 6995:7015]  # pozo
# crop = img[1900:2500, 8710:8750]  # vias

hsv_pozos = cv.cvtColor(img_pozos, cv.COLOR_BGR2HSV)
hsv_via1 = cv.cvtColor(img_via1, cv.COLOR_BGR2HSV)
hsv_via2 = cv.cvtColor(img_via2, cv.COLOR_BGR2HSV)

# mask color #00FF00 Green1
pozo_mask = cv.inRange(hsv_pozos, (45, 195, 245), (48, 198, 252))
via1_mask = cv.inRange(hsv_via1, (45, 195, 245), (48, 198, 252))
via2_mask = cv.inRange(hsv_via2, (45, 195, 245), (48, 198, 252))

# dilatation_size = 5
# dilation_shape = cv.MORPH_ELLIPSE
# element = cv.getStructuringElement(dilation_shape, (2 * dilatation_size + 1, 2 * dilatation_size + 1),
#                                     (dilatation_size, dilatation_size))
# pozo_mask = cv.erode(pozo_mask, element)
# pozo_mask = cv.dilate(pozo_mask, element)
#
# contours, _ = cv.findContours(pozo_mask, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE)
# # cv.drawContours(mask, contours, -1, 255, 30)
# font = cv.FONT_HERSHEY_SIMPLEX
#
# for i, c in enumerate(contours):
#     x, y = c.mean(axis=0)[0]
#     cv.circle(img, (int(x), int(y)), 25, (0, 0, 0), -1)
#     cv.putText(img, str(i),
#                (int(x + 50), int(y + 50)),
#                font, 5, (0, 0, 0), 10, cv.FILLED)

# show processed image
cv.imshow('map', via2_mask + via1_mask)
cv.waitKey(0)



import os
from tqdm import tqdm

import googlemaps
import polars as pl

from chaparral.mapa.convenciones import CLUSTER2MAPA, POZO2CLUSTER

API_KEY = os.environ['GOOGLE_API_KEY']
gmaps = googlemaps.Client(key=API_KEY)
objetos = pl.read_ndjson('data/objetos_mapa/objetos.ndjson')
obj_coords = pl.read_ndjson('data/objetos_mapa/obj_coords.ndjson')


def get_vecinos(df, cluster, iter=0) -> set:
    """Encuentra los clusteres vecinos al target cluster.
    Argument:
        df: dataframe con informacion de los objetos cargados en el mapa.
        cluster: el cluster target.
    """
    if cluster not in df['nombre']:
        raise ValueError(f'Objeto {cluster} not in list')
    cl_info = df.select(['nombre', 'vecinos']).filter(pl.col('nombre') == cluster).row(0, named=True)
    vecinos = cl_info['vecinos']
    results = []
    for vecino in vecinos:
        if 'cluster' not in vecino and 'bodega' not in vecino:
            if iter < 2:
                segundos = get_vecinos(df, vecino, iter + 1)
                results.extend([v for v in segundos if 'cluster' in v])
        else:
            results.append(vecino)
    return set([v for v in results if v != cluster])


results = []
for objeto in objetos.iter_rows(named=True):
    tipo = objeto['tipo']
    if 'CLUSTER' in tipo:
        nombre = CLUSTER2MAPA[objeto['nombre'].upper()]
    elif 'BODEGA' in tipo:
        nombre = 'bodega_00'
    else:
        continue
    codigo = objeto['nombre']
    pozos = [k for k, v in POZO2CLUSTER.items() if v == nombre]
    vecinos = [CLUSTER2MAPA[vecino.upper()] if vecino.startswith('cluster') else CLUSTER2MAPA[vecino]
               for vecino in get_vecinos(objetos, codigo)]
    coords = obj_coords.filter(pl.col('nombre') == nombre).select(['lat', 'lon'])[0].to_dict()
    row = dict(
        nombre=nombre,
        codigo=codigo,
        tipo=tipo,
        vecinos=vecinos,
        pozos=pozos,
        coords=(coords['lat'][0], coords['lon'][0]),
    )
    results.append(row)

results = pl.DataFrame(results)

final = []
pbar = tqdm(total=len(results), desc='Adquiriendo aristas')
for row in results.iter_rows(named=True):
    new_vecinos = []
    for vecino in row['vecinos']:
        vecino_coords = results.filter(pl.col('nombre') == vecino).row(0, named=True)['coords']
        request = gmaps.distance_matrix(
            tuple(row['coords']),
            tuple(vecino_coords),
            mode='driving',
            units='metric'
        )
        if request['rows'][0]['elements'][0]['distance']['value'] == 0:
            print((row['nombre'], row['coords']), (vecino, vecino_coords), request)
        vecino_row = dict(
            nombre=vecino,
            dist=request['rows'][0]['elements'][0]['distance']['value'],
            tiempo=request['rows'][0]['elements'][0]['duration']['value'],
        )
        new_vecinos.append(vecino_row)
    row['vecinos'] = new_vecinos
    final.append(row)
    pbar.update(1)

final = pl.DataFrame(final)
final.write_ndjson('data/objetos_mapa/objetos_api.ndjson')
print(final)



'''
© Copyright ArdillaByte Inc. 2023
-----------------------------------------------
NN arquitectures
-----------------------------------------------
'''
import torch
from typing import List, Union
from copy import deepcopy



class MLP(torch.nn.Module):
    '''
    A Multi-layer Perceptron
    '''
    def __init__(self, 
                 sizes:List[int], 
                 intermediate_activation_function:any,
                 last_activation_function:Union[None, any]):
        """
        Args:
            sizes (list): list with the sizes of the layers. 
                          The convention is sizes[0] is the size of the input layer
                          and sizes[-1] is the size of the output layer.
            last_activation_function (an activation function)
        """
        super().__init__()
        assert(len(sizes) > 1)
        self.sizes = sizes
        self.intermediate_activation_function = intermediate_activation_function
        self.last_activation_function = last_activation_function
        # -------------------------------------
        # Defining the layers
        # -------------------------------------
        self.model = torch.nn.Sequential()
        # # Apply batch normalization to input
        # self.model.append(torch.nn.BatchNorm1d(sizes[0]))
        for i in range(len(sizes) - 1):
            n_from = sizes[i]
            n_to = sizes[i+1]
            # print(f'Creando capa lineal {n_from} x {n_to}')
            self.model.append(torch.nn.Linear(n_from, n_to))
            # if i < len(sizes) - 3:
            #     # print('Incluyendo layer norm')
            #     self.model.append(torch.nn.LayerNorm(n_to))
            if i < len(sizes) - 2:
                # print('Incluyendo activación')
                self.model.append(self.intermediate_activation_function)
        if self.last_activation_function is not None:
            self.model.append(self.last_activation_function)

    def forward(self, x_in):
        """The forward pass of the network        
        Args:
            x_in (torch.Tensor): an input data tensor. 
        Returns:
            the resulting tensor.
        """
        # Run the input through layers 
        # if len(x_in.shape) == 1:
        #     x_in = x_in.unsqueeze(dim=0)
        return self.model(x_in)
    


'''
© Copyright ArdillaByte Inc. 2023

-----------------------------------------------
Class to run, renderize, train and test agents
over environments.
-----------------------------------------------
'''
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
from copy import deepcopy
from typing import Dict
from os import path
from pathlib import Path
import sys
import chaparral.modelos.entorno as E
from chaparral.utils.interaction import Episode, Experiment
import chaparral.agents.agentsNN as ApproxNN
from chaparral.utils.plot_utils import Plot

SCRIPT_PATH = Path.cwd() / Path('..').resolve()

OWN_ENV_LIST = ['ABC', 'GridworldEnv', 'PitLaberynth']


class Performer :
    '''
    Class to train and run an agent in an environment.
    '''
    def __init__(self,\
                env_name:str,\
                env_parameters:Dict[str, any],\
                state_interpreter:any,\
                agent_name:str,\
                agent_parameters:Dict[str,any]
                ) -> None:
        #--------------------------------
        # Environment metadata
        #--------------------------------
        self.env_name = env_name
        assert(isinstance(env_parameters, dict))
        self.env_parameters = env_parameters
        # to check if environment is given to use
        self.use_environment = False
        self.state_interpreter = state_interpreter
        #--------------------------------
        # Agent metadata
        #--------------------------------
        if '.' in agent_name:
          agent_class, agent_name = agent_name.split('.')
        else:           
          agent_class = 'TableAgents'
        self.agent_name = agent_name
        self.agent_class = agent_class
        self.agent_parameters = agent_parameters
        if 'deep' in agent_parameters.keys():
          self.deep = agent_parameters['deep']
        else:
          self.deep = False
        self.dict_paths = self.consolidate_folders()
        self.data = None

    def consolidate_folders(self) -> Dict[str,Path]:
        """Create folders to store model files"""
        dict_paths = dict()
        dict_paths['file_name'] = f'{self.agent_name}_in_{self.env_name}'
        dict_paths['image_folder'] = SCRIPT_PATH / Path('images', dict_paths['file_name'])
        dict_paths['image_folder'].mkdir(parents=True, exist_ok=True)
        dict_paths['data_folder'] = SCRIPT_PATH / Path('data', dict_paths['file_name'])
        dict_paths['data_folder'].mkdir(parents=True, exist_ok=True)
        dict_paths['model_folder'] = SCRIPT_PATH / Path('models', dict_paths['file_name'])
        dict_paths['model_folder'].mkdir(parents=True, exist_ok=True)
        dict_paths['video_folder'] = SCRIPT_PATH / Path('videos', dict_paths['file_name'])
        dict_paths['video_folder'].mkdir(parents=True, exist_ok=True)
        self.extension = '.pt' if self.deep else '.json'
        dict_paths['file_model'] = path.join(dict_paths['model_folder'], f'{dict_paths["file_name"]}{self.extension}')
        dict_paths['file_csv'] = path.join(dict_paths['data_folder'], f'{dict_paths["file_name"]}.csv')
        dict_paths['file_png'] = path.join(dict_paths['image_folder'], f'{dict_paths["file_name"]}.png')
        dict_paths['file_losses'] = path.join(dict_paths['image_folder'], f'{dict_paths["file_name"]}_losses.png')
        dict_paths['file_test'] = path.join(dict_paths['image_folder'], f'{dict_paths["file_name"]}_test.png')
        dict_paths['file_test_csv'] = path.join(dict_paths['data_folder'], f'{dict_paths["file_name"]}_test_csv.csv')
        dict_paths['file_compare_hist'] = path.join(dict_paths['image_folder'], f'comparison_hist.png')
        dict_paths['file_compare_rew'] = path.join(dict_paths['image_folder'], f'comparison_rew.png')
        return dict_paths

    def load_env(self, render_mode:str) -> None:
        '''
        Load environment. Render mode is different 
        for training (None) than for running (rgb_array). Render
        mode can only be set when instantiating the environment.
        '''
        # check if environment is given to be used
        if self.use_environment:
           return None
        if self.env_name in OWN_ENV_LIST:
            if self.env_parameters is not None:
              exec(f'self.environment = E.{self.env_name}(**self.env_parameters)')              
            else:
              exec(f'self.environment = E.{self.env_name}()')
            self.environment.render_mode = render_mode
        else:
          try:
            self.environment = gym.make(
                self.env_name, 
                render_mode=render_mode,
                **self.env_parameters
            )
          except:
            raise Exception(f'Environment {self.env_name} unknown.')

    def use_env(self, env:any) -> None:
       """Takes an environment for use in interaction"""
       assert(hasattr(env, 'step'))
       assert(hasattr(env, 'reset'))
       assert(hasattr(env, 'render'))
       assert(hasattr(env, 'render_mode'))
       assert(env.render_mode == 'rgb_array')
       self.environment = env
       self.use_environment = True

    def load_agent(self, from_file:bool=False):
        '''
        Load agent from name
        '''
        if self.agent_class == 'TabularAgents':
          line = f'self.agent = TableA.{self.agent_name}(self.agent_parameters)'
          self.deep = False
        elif self.agent_class == 'agentsCS':
          line = f'self.agent = ApproxQ.{self.agent_name}(self.agent_parameters)'
          self.deep = False
        elif self.agent_class == 'agentsNN':
          line = f'self.agent = ApproxNN.{self.agent_name}(self.agent_parameters)'
          self.deep = True
        elif self.agent_class == 'agentsPG':
          line = f'self.agent = ApproxP.{self.agent_name}(self.agent_parameters)'
        else:
           raise Exception(f'Agent class {self.agent_class} is unknown!')
        exec(line)
        if from_file:
          print(f'Loading agent from {self.dict_paths["file_model"]}')
          self.agent.load(file=self.dict_paths['file_model'])
    
    def save_agent(self):
        try:
            self.agent.save(file=self.dict_paths['file_model'])
        except Exception as e:
            print('\n\tAn error occurred:\n\t', e,'\n')
            pass

    def shutdown_agent_exploration(self) -> (float, np.ndarray):
        backup_epsilon = deepcopy(self.agent.epsilon)
        self.agent.epsilon = 0
        if hasattr(self.agent, 'policy'):
          backup_policy = deepcopy(self.agent.policy)
          for s in range(self.agent.nS):
            self.agent.update_policy(s)
        else:
          backup_policy = None
        return backup_epsilon, backup_policy

    def run(self, 
            from_file:bool=False,
            no_exploration:bool=False,
            visual:bool=True, 
            to_video:bool=False,
            sleep_time:float=0.3,
            num_rounds:int=200):
        '''
        Run the agent on the environment and displays the behavior.
        Agent does not learn.
        Input:
          - from_file (bool), if true, attemts to load the agent from file
          - no_exploration (bool), if true, makes epsilon = 0
          - visual (bool),
            True: displays the environment as in a video using environment render
            False: displays the behavioral data in the terminal step by step
          - to_video (bool),
            True: saves the rendering to a video file 
            False: displays the environment as in a video using environment render
          - sleep_time (float), determines the speed of the renderization
          - num_rounds (int), number of rounds to display
        '''
        # Load agent from name
        self.load_agent(from_file=from_file)
        # self.agent.debug = True # Uncomment for debugging
        if no_exploration:
          backup_epsilon, backup_policy = self.shutdown_agent_exploration()
        if visual and not to_video:
          '''
          To display the environment as in a video
          '''
          # Create environment
          # self.load_env(render_mode='human')
          self.load_env(render_mode='rgb_array')
          try:
            self.environment._max_episode_steps = num_rounds
          except:
             pass
          # Create episode
          episode = Episode(
              environment=self.environment,\
              env_name=self.env_name,\
              agent=self.agent,\
              model_name=self.agent_name,\
              num_rounds=num_rounds,\
              state_interpreter=self.state_interpreter
          )
          episode.sleep_time = sleep_time
          episode.renderize(to_video=False)
        elif to_video:
          '''
          To save to a video file
          '''
          # Create environment
          self.load_env(render_mode='rgb_array')
          try:
            self.environment._max_episode_steps = num_rounds
          except:
             pass
          # Create episode
          episode = Episode(
              environment=self.environment,\
              env_name=self.env_name,\
              agent=self.agent,\
              model_name=self.agent_name,\
              num_rounds=num_rounds,\
              state_interpreter=self.state_interpreter
          )
          episode.renderize(
              to_video=True,
              file=self.dict_paths['video_folder']
          )
        else:
          '''
          To display data information in the terminal
          '''
          # Create environment
          self.load_env(render_mode=None)
          try:
            self.environment._max_episode_steps = num_rounds
          except:
             pass
          # Create episode
          episode = Episode(
              environment=self.environment,\
              env_name=self.env_name,\
              agent=self.agent,\
              model_name=self.agent_name,\
              num_rounds=num_rounds,\
              state_interpreter=self.state_interpreter
          )
          df = episode.run(verbose=4, learn=False)
          self.data = df
        print('Number of rounds:', len(episode.agent.rewards) - 1)
        print('Total reward:', np.nansum(episode.agent.rewards))
        if no_exploration:
          self.agent.epsilon = backup_epsilon
          self.agent.policy = backup_policy
            
    def train(
            self, 
            num_rounds:int=200, 
            num_episodes:int=500, 
            from_file:bool=False
          ) -> None:
        '''
        Trains agent.
        '''
        # Load agent from name
        self.load_agent(from_file=from_file)
        # Create environment
        self.load_env(render_mode=None)
        try:
          self.environment._max_episode_steps = num_rounds
        except:
            pass
        # Create episode
        episode = Episode(
            environment=self.environment,\
            env_name=self.env_name,\
            agent=self.agent,\
            model_name=self.agent_name,\
            num_rounds=num_rounds,\
            state_interpreter=self.state_interpreter
        )
        # lengths = f'#states:{len(episode.agent.states)} -- #actions:{len(episode.agent.actions)} -- #rewards:{len(episode.agent.rewards)} -- #dones:{len(episode.agent.dones)}'
        # print('/>', lengths)
        # Train agent
        df = episode.simulate(
            num_episodes=num_episodes, 
            file=self.dict_paths['file_csv']
        )
        self.data = df
        print(f'Data saved to {self.dict_paths["file_csv"]}')
        # Save agent to file
        self.save_agent()
        print(f'Agent saved to {self.dict_paths["file_model"]}')
        # Plot results
        p =  Plot(df)
        if num_episodes == 1:
          p.plot_round_reward(file=self.dict_paths['file_png'])    
        else:
          p.plot_rewards(file=self.dict_paths['file_png']) 
        print(f'Plot saved to {self.dict_paths["file_png"]}')
        # Save losses if agent uses NN
        if hasattr(self.agent, 'NN'):
          if hasattr(self.agent.NN, 'losses'):
            losses = self.agent.NN.losses
            fig, ax = plt.subplots(figsize=(4,3.5))
            ax = sns.lineplot(x=range(len(losses)), y=losses)
            ax.set_xlabel("Epoch",fontsize=14)
            ax.set_ylabel("Loss",fontsize=14)
            plt.savefig(self.dict_paths['file_losses'], dpi=300, bbox_inches="tight")
        elif hasattr(self.agent, 'policy'):
          if hasattr(self.agent.policy, 'losses'):
            losses = self.agent.policy.losses
            fig, ax = plt.subplots(figsize=(4,3.5))
            ax = sns.lineplot(x=range(len(losses)), y=losses)
            ax.set_xlabel("Epoch",fontsize=14)
            ax.set_ylabel("Loss",fontsize=14)
            plt.savefig(self.dict_paths['file_losses'], dpi=300, bbox_inches="tight")

    def test(self, 
             no_exploration:bool=True,
             from_file:bool=True, 
             num_rounds:int=200, 
             num_episodes:int=100):
        '''
        Test the trained agent.
        '''
        # Load agent from name
        self.load_agent(from_file=from_file)
        if no_exploration:
          # Shutdown exploration
          backup_epsilon, backup_policy = self.shutdown_agent_exploration()
        # Create environment
        self.load_env(render_mode=None)
        try:
          self.environment._max_episode_steps = num_rounds
        except:
            pass
        # Create episode
        episode = Episode(
            environment=self.environment,\
            env_name=self.env_name,\
            agent=self.agent,\
            model_name=self.agent_name,\
            num_rounds=num_rounds,\
            state_interpreter=self.state_interpreter
        )
        # Run simulation
        df = episode.simulate(
            num_episodes=num_episodes, 
            learn=False
        )
        self.data = df
        df.to_csv(self.dict_paths['file_test_csv'])
        print(f'Data saved to {self.dict_paths["file_test_csv"]}')
        # Plot results
        p = Plot(df)
        p.plot_histogram_rewards(self.dict_paths['file_test'])
        print(f'Plot saved to {self.dict_paths["file_test"]}')
        if no_exploration:
          self.agent.epsilon = backup_epsilon
          if hasattr(self.agent, 'policy'):
            self.agent.policy = backup_policy
         
    def sweep(self, 
              parameter:str, 
              values:list, 
              num_rounds:int=200, 
              num_episodes:int=100,
              num_simulations:int=10):
        '''
        Runs a sweep over the specified parameter 
        with the specified values.
        '''
        # Load agent from name
        self.load_agent()
        # Creates environment
        self.load_env(render_mode=None)
        # Creates experiment
        experiment = Experiment(
            environment=self.environment,\
            env_name=self.env_name,\
            num_rounds=num_rounds,\
            num_episodes=num_episodes,\
            num_simulations=num_simulations,\
            state_interpreter=self.state_interpreter
        )
        # Run sweep
        experiment.run_sweep1(
            agent=self.agent, \
            name=self.agent_name, \
            parameter=parameter, \
            values=values, \
            measures=['reward']
        )
        self.data = experiment.data
        # Plot results
        p = Plot(experiment.data)
        print('Plotting...')
        p.plot_rewards(self.dict_paths['file_compare_rew'])
        print(f'Plot saved to {self.dict_paths["file_compare_rew"]}')
        
    def compare_test(self, 
                     agent_vs_name:str,
                     agent_vs_parameters:Dict,
                     num_rounds:int=200, 
                     num_episodes:int=100):
        '''
        Runs a comparison of two agents
        over an environment.
        Agents are loaded from file.
        '''
        # Load agent 1
        self.load_agent(from_file=True)
        self.shutdown_agent_exploration()
        agent1 = deepcopy(self.agent)
        # Load vs agent
        backup_agent_name = self.agent_name
        backup_agent_parameters = deepcopy(self.agent_parameters)
        self.agent_name = agent_vs_name
        self.agent_parameters = agent_vs_parameters
        self.consolidate_folders()
        try:
          self.load_agent(from_file=True)
        except Exception as e:
          print(e)
          print(f'An agent of class {agent_vs_name} is required.\nRun a performer on such an agent first.') 
        self.shutdown_agent_exploration()
        agent2 = deepcopy(self.agent)
        self.agent_name = backup_agent_name
        self.agent_parameters = backup_agent_parameters
        self.consolidate_folders()
        # Create environment
        self.load_env(render_mode=None)
        # Create experiment
        experiment = Experiment(
            environment=self.environment,\
            env_name=self.env_name,\
            num_rounds=num_rounds,\
            num_episodes=num_episodes,\
            num_simulations=1,\
            state_interpreter=self.state_interpreter
        )
        # Run sweep
        experiment.run_experiment(
            agents=[agent1, agent2], \
            names=[self.agent_name, agent_vs_name], \
            measures=['hist_reward'],\
            learn=False
        )
        self.data = experiment.data
        # Plot results
        p = Plot(experiment.data)
        print('Plotting...')
        p.plot_histogram_rewards(self.dict_paths['file_compare_hist'])
        print(f'Plot saved to {self.dict_paths["file_compare_hist"]}')




from pathlib import Path
import pandas as pd
from typing import List, Dict, Optional, Tuple, Union
import warnings
import numpy as np
import json
import pickle
from copy import deepcopy

from chaparral.modelos.clases import (
	Cluster, 
	Camion, 
	QuimicoEn, 
	Pozo
)
from chaparral.modelos.entorno import CampoPozos
from chaparral.main.cargar_inventario import estandarizar_nombre
# from chaparral.mapa.convenciones import POZO2CLUSTER 
from chaparral.modelos.constantes import (
	ESCALA_VELOCIDAD_CONSUMO, 
	ESCALA_VELOCIDAD_CAMION,
	VALOR_CRITICO_CARGA
)


class Auditor :
	"""Clase que revisa si el CampoPozos tiene los requerimientos necesarios"""

	def __init__(self) -> None:
		self.errores_log = list()
		self.directorio_log = Path.cwd() / Path('..').resolve() / Path('data', 'logs')
		self.directorio_log.mkdir(parents=False, exist_ok=True)

	def auditar_quimicos(
				self,
				pozos:List[Pozo]
			) -> List[Pozo]:
		pozos_a_eliminar = list()
		for pozo in pozos:
			if len(pozo.quimicos) == 0:
				mensaje_error = f'Pozo {pozo.nombre} no tiene quimicos'
				warnings.warn('Advertencia de error en lectura de datos cluster.') 
				self.errores_log.append(mensaje_error)
				pozos_a_eliminar.append(pozo.nombre)
		return [pozo for pozo in pozos if pozo.nombre not in pozos_a_eliminar]

	def auditar_integridad_clusters(
				self,
				clusters:List[Cluster]
			) -> Tuple[List[Cluster], bool]:
		assert(clusters[0].nombre == 'bodega_00'), f'Cluster 0 debe ser bodega_00, no {clusters[0].nombre}'
		clusters_a_eliminar = list()
		for cluster in clusters:
			# Verifica los químicos de los pozos en el cluster
			cluster.pozos = self.auditar_quimicos(cluster.pozos)
			# Verifica el nombre del cluster
			if not isinstance(cluster.nombre, str):
				mensaje_error = f'Cluster {cluster.nombre} sin nombre de tipo adecuado'
				warnings.warn('Advertencia de error en químicos asociados a cluster.') 
				self.errores_log.append(mensaje_error)
				clusters_a_eliminar.append(cluster.nombre)
			# Verifica las coordenadas del cluster
			if not isinstance(cluster.coordenadas, list) and len(cluster.coordenadas) == 2:
				mensaje_error = f'Cluster {cluster.nombre} sin coordenadas de tipo adecuado'
				warnings.warn('Advertencia de error en químicos asociados a cluster.') 
				self.errores_log.append(mensaje_error)
				clusters_a_eliminar.append(cluster.nombre)
			else:
				x, y = cluster.coordenadas[0], cluster.coordenadas[1]	
				if not (isinstance(x, float) and isinstance(y, float)):
					mensaje_error = f'Cluster {cluster.nombre} sin coordenadas de tipo adecuado'
					warnings.warn('Advertencia de error en químicos asociados a cluster.') 
					self.errores_log.append(mensaje_error)
					clusters_a_eliminar.append(cluster.nombre)		
			# Verifica que haya pozos en el cluster	
			if not len(cluster.pozos) > 0 and cluster.nombre != 'bodega_00':
				mensaje_error =  f'Cluster {cluster.nombre} sin pozos'
				warnings.warn('Advertencia de error en químicos asociados a cluster.') 
				self.errores_log.append(mensaje_error)
				clusters_a_eliminar.append(cluster.nombre)
			# if not len(cluster.vecinos) > 0:
			# 	mensaje_error = f'Cluster {cluster.nombre} sin vecinos'
			# 	warnings.warn('Advertencia de error en químicos asociados a cluster.') 
			# 	self.errores_log['clusters & quimicos'].append(mensaje_error)
			# 	clusters_a_eliminar.append(cluster.nombre)			
		if len(clusters_a_eliminar) > 0:
			flag = True
		else:
			flag = False
		clusters = [cluster for cluster in clusters if cluster.nombre not in clusters_a_eliminar]
		return clusters, flag
	
	def auditar_clusters(
				self,
				clusters:List[Cluster]
		) -> List[Cluster]:
		flag = True
		while flag:
			clusters, flag1 = self.auditar_vecinos_en_clusters(clusters)
			clusters, flag2 = self.auditar_integridad_clusters(clusters)
			flag = flag1 or flag2
		return clusters
	
	def auditar_vecinos_en_clusters(
				self,
				clusters:List[Cluster]
			) -> Tuple[List[Cluster], bool]:
		"""Revisa que los vecinos de cada cluster estén incluidos"""
		nombres_clusters = [cluster.nombre for cluster in clusters]
		flags = list()
		for cluster in clusters:
			cluster.vecinos = [vecino for vecino in cluster.vecinos if vecino in nombres_clusters]
			if len(cluster.vecinos) == 0:
				flags.append(True)
			else:
				flags.append(False)
		return clusters, np.any(flags)

	def auditar_cluster_a_pozos(self) -> None:
		"""
		Revisa que la información de cluster a puntos coincida
		con los clusters dados como argumento.
		"""		
		# Creamos ambas listas de clusters
		clusters_desde_argumento = set([cluster.nombre for cluster in self.clusters])
		clusters_desde_argumento.add('bodega_00')
		clusters_en_cluster_a_puntos = set(self.cluster_a_pozos.keys())
		clusters_en_cluster_a_puntos.add('bodega_00')
		# Encontramos diferencias
		errores1 = clusters_desde_argumento.difference(clusters_en_cluster_a_puntos)
		errores2 = clusters_en_cluster_a_puntos.difference(clusters_desde_argumento)
		# Actualizamos el log de errores
		if len(errores1) > 0:
			warnings.warn('Advertencia de error en lectura de datos cluster.') 
			self.errores_log.append(f'clusters sin datos de puntos:\n{errores1}')
		if len(errores2) > 0:
			warnings.warn('Advertencia de error en lectura de datos cluster.') 
			self.errores_log.append(f'clusters con puntos sin datos de rutas:\n{errores2}')
		# Dejamos solo los clusters con toda la información
		clusters_a_usar = clusters_desde_argumento.intersection(clusters_en_cluster_a_puntos)
		self.clusters = [cluster for cluster in self.clusters if cluster.nombre in clusters_a_usar]

	def guardar(self) -> None:
		# Guardamos el registro de errores
		json_object = json.dumps(self.errores_log, indent=4)
		nombre_archivo_log = self.directorio_log / Path('auditor_log.json')
		with open(nombre_archivo_log, "w") as outfile:
			outfile.write(json_object)


class LectorQuimicos :
	"""Lee información de quimicos y sus cantidades en los clusters."""

	def __init__(
				self, 
				nombre_archivo:str
			) -> Dict[str, Dict[str, float]]:
		self.nombre_archivo = nombre_archivo
		self.data = None
		self.verbose = False
		# Inicializa el registro de quimicos
		self.quimicos = list()
		self.nombres_quimicos = list()
		self.dict_quimicos = {quimico.nombre:p for p, quimico in enumerate(self.quimicos)}
		# Inicializa el registro de errores
		self.errores_log = list()
		# # Genera diccionario de cluster a puntos
		# self.cluster_a_pozos = self.crear_cluster_a_pozos()
		# # Compara datos de puntos a cluster con clusters recibidos como argumento
		# self.auditar_cluster_a_pozos()

	def leer_datos(self) -> None:
		"""Lee datos de archivo y crea el dataframe de info."""
		# Lee datos desde archivo
		self.data = pd.read_json(self.nombre_archivo, lines=True)
		if self.verbose:
			print(f'Datos leídos con éxito desde el archivo {self.nombre_archivo}')

	def actualizar_clusters(
				self,
				clusters:List[Cluster]
			) -> Tuple[List[Cluster], List[str]]:
		"""Actualiza los clusters dados con la info."""
		# Lee los datos desde archivo
		self.leer_datos()
		# Inicializa listado de clusters sin datos
		clusters_sin_datos = list()
		# Itera sobre clusters
		for cluster in clusters[1:]: # Omitir Bodega
			# Itera sobre pozo en cluster
			for pozo in cluster.pozos:
				lista_pozos = self.data.punto.unique().tolist()
				# if self.verbose:
				# 	print(f'Actualizando químicos de pozo {pozo.nombre}')
				if pozo.nombre in lista_pozos:
					quimicos_en_cluster = self.data.groupby('punto').get_group(pozo.nombre)
					quimicos = list()
					# Itera sobre químico en pozo
					for nombre_quimico, grp in quimicos_en_cluster.groupby('quimico'):
						quimico = self.crear_quimico(
							nombre_quimico=nombre_quimico, 
							datos=grp
						)
						quimicos.append(quimico)
					pozo.quimicos = quimicos
					self.quimicos += quimicos
				else:
					mensaje_error = f'No parece haber datos del pozo {pozo}.'
					self.errores_log.append(mensaje_error)
			# pozos = self.auditar_quimicos(pozos)
		if self.verbose:
			print('Clusters actualizados con datos de los químicos.')
		# Limpia registros clusters
		clusters = [cluster for cluster in clusters if cluster.nombre not in clusters_sin_datos]
		# Limpia registro químicos
		nombres_quimicos = list(set([quimico.nombre for quimico in self.quimicos]))
		self.quimicos = [
			QuimicoEn(
				nombre=nombre,
				unidades='Gl',
				nivel_actual=0,
				velocidad_consumo=0
			) for nombre in nombres_quimicos
		]
		return clusters, self.errores_log
	
	def crear_quimico(
				self,
				nombre_quimico:str,
				datos:pd.DataFrame
			) -> QuimicoEn:
		nivel_actual = datos.nivel_actual.values.tolist()[0]
		velocidad_consumo = datos.velocidad_consumo.values.tolist()[0]
		quimico = QuimicoEn(
			nombre=nombre_quimico,
			unidades='Gl',
			nivel_actual=nivel_actual,
			velocidad_consumo=velocidad_consumo * ESCALA_VELOCIDAD_CONSUMO
		)
		return quimico
	
	# def crear_cluster_a_pozos(self) -> Dict[str, str]:
	# 	"""Crea el diccionario que relaciona un cluster con sus pozos/puntos"""
	# 	df = pd.DataFrame({'pozo':POZO2CLUSTER.keys(), 'cluster':POZO2CLUSTER.values()})
	# 	cluster_a_pozos = dict()
	# 	for cluster, grp in df.groupby('cluster'):
	# 		if cluster != 'bodega_00':
	# 			pozos = grp.pozo.unique().tolist()
	# 			cluster_a_pozos[cluster] = pozos
	# 			if len(pozos) == 0:
	# 				mensaje_error = f'Cluster {cluster} no tiene pozos'
	# 				warnings.warn('Advertencia de error en lectura de datos cluster.') 
	# 				self.errores_log.append(mensaje_error)
	# 	return cluster_a_pozos


class CreadorCampoPozos :
	"""
	Crea la clase CampoPozos con todos sus componentes.
	"""

	def __init__(
				self,
				nombre_archivo_clusters:str,
				nombre_archivo_quimicos:str,
				nombre_archivo_rutas:str
			) -> CampoPozos:
		# -------------------------------
		# Registros de directorios
		# -------------------------------
		self.directorio = Path.cwd() / Path('..').resolve() / Path('data', 'objetos_mapa')
		self.directorio.mkdir(parents=False, exist_ok=True)
		self.directorio_log = Path.cwd() / Path('..').resolve() / Path('data', 'logs')
		self.directorio_log.mkdir(parents=False, exist_ok=True)
		# -------------------------------
		# Registros de archivos
		# -------------------------------
		self.nombre_archivo_clusters = nombre_archivo_clusters
		self.nombre_archivo_quimicos = nombre_archivo_quimicos
		self.nombre_archivo_rutas = nombre_archivo_rutas
		# -------------------------------
		# Registro de datos
		# -------------------------------
		self.data = None
		self.nombres_clusters = list()
		self.verbose = False
		self.errors_log = {
			'clusters':list(),
			'clusters & quimicos':list()
		}

	def crear_campo(self) -> None:
		# -------------------------------
		# Carga los datos desde archivo
		# -------------------------------
		self.cargar_datos()
		# -------------------------------
		# Crea los clusters y los quimicos
		# -------------------------------
		clusters, quimicos = self.generar_clusters_y_quimicos()
		# -------------------------------
		# Auditar clusters
		# -------------------------------
		auditor = Auditor()
		clusters, flag = auditor.auditar_integridad_clusters(clusters)
		nombres_clusters = [cluster.nombre for cluster in clusters]
		auditor.guardar()
		if self.verbose:
			print(f'Log de auditor guardado')
		# -------------------------------
		# Carga las rutas
		# -------------------------------
		datos_rutas = self.cargar_rutas()
		# -------------------------------
		# Crea el camion
		# -------------------------------
		# Inicializa la carga del camión
		print('')
		print('Químicos a camión:')
		for quimico in quimicos:
			quimico.nivel_actual = VALOR_CRITICO_CARGA * quimico.velocidad_consumo
			quimico.nivel_inicial = quimico.nivel_actual
			print(quimico.nombre, quimico.nivel_actual)
		camion1 = Camion(
			nombre='C1',
			capacidad=100,
			carga_quimicos=quimicos,
			ubicacion=0,
			ruta=None,
			nombres_clusters=nombres_clusters
		)
		# -------------------------------
		# Crea el campo de pozos
		# -------------------------------
		campo = CampoPozos(
			clusters=clusters,
			datos_rutas=datos_rutas,
			camiones=[camion1],
		)
		return campo

	def cargar_datos(self) -> None:
		ruta_archivo = self.directorio / Path(self.nombre_archivo_clusters)
		with open(ruta_archivo) as f:
			self.data = pd.read_json(f, lines=True)
		if self.verbose:
			print('')
			print(f'Datos cargados con éxito desde archivo.\n\t({ruta_archivo})')

	def cargar_rutas(self) -> None:
		ruta_archivo = self.directorio / Path(self.nombre_archivo_rutas)
		with open(ruta_archivo) as f:
			data = json.load(f)
		return data

	def generar_clusters_y_quimicos(self) -> List[Cluster]:
		# Creamos los clusters
		if self.verbose:
			print('')
			print('Creando clusters...')
		clusters = self.crear_clusters()
		# Cargamos la información de químicos
		if self.verbose:
			print('')
			print('Actualizando información de químicos en clusters...')
		nombre_archivo=self.directorio / Path(self.nombre_archivo_quimicos)
		lector = LectorQuimicos(nombre_archivo)
		lector.verbose = self.verbose
		clusters, errores = lector.actualizar_clusters(clusters)
		self.errors_log['quimicos'] = errores
		# Guardamos el registro de errores
		json_object = json.dumps(self.errors_log, indent=4)
		nombre_archivo_log = self.directorio_log / Path('creador_log.json')
		with open(nombre_archivo_log, "w") as outfile:
			outfile.write(json_object)
		if self.verbose:
			print(f'Log de errores guardado en {nombre_archivo_log}')
		return clusters, deepcopy(lector.quimicos)

	def crear_clusters(self) -> Tuple[List[Cluster], Dict[str,int]] :
		# Inicializamos la lista de clusters
		clusters = list()
		data_bodega = self.data[self.data.tipo == 'BODEGA']
		bodega = self.crear_cluster(
			data_bodega, 
			bodega=True
		)
		if bodega is not None:
			clusters.append(bodega)
			if self.verbose:
				print('')
				print('\tBodega creada')
		else:
			print('Advertencia: Bodega no se creó. Revisar log.')
		data_clusters = self.data[self.data.tipo == 'CLUSTER']
		if self.verbose:
			print('')
			print(f'\tCreando {data_clusters.shape[0]} clusters...')
		for indice, data_cluster in data_clusters.iterrows():
			cluster = self.crear_cluster(data_cluster)
			clusters.append(cluster)
		return clusters

	def crear_cluster(
			self,
			data_cluster:pd.DataFrame,
			bodega:bool=False,
		) -> Cluster:
		"""Crea el cluster a partir de los datos"""
		# Extraemos el nombre
		if bodega:
			nombre = 'bodega_00'
		else:
			nombre = str(data_cluster.nombre)
		# Incluimos el nombre en la lista
		if nombre not in self.nombres_clusters:
			self.nombres_clusters.append(nombre)
		# Extraemos las coordenadas
		coordenadas = data_cluster.coords
		if isinstance(coordenadas, pd.Series):
			coordenadas = coordenadas.values[0]
		elif isinstance(coordenadas, np.ndarray):
			coordenadas = coordenadas.tolist()
		# Extraer información de pozos
		pozos = list()
		for pozo_ in data_cluster.pozos:
			pozo = Pozo(
				nombre=pozo_,
				quimicos=list()
			)
			pozos.append(pozo)
		# # Extraemos toda la información de los vecinos
		# data_vecinos = list(data_cluster.vecinos)
		# if bodega:
		# 	data_vecinos = data_vecinos[0]
		# 	nombre = 'bodega_00'
		# # Procesamos la información de las rutas a vecinos
		# try:
		# 	self.agregar_rutas(
		# 		nombre_cluster=nombre, 
		# 		data_vecinos=data_vecinos
		# 	)
		# 	if self.verbose:
		# 		print(f'\t\tRutas con vecinos actualizadas para cluster {nombre}...')
		# except Exception as e:
		# 	error_message = f'Error con cluster {nombre}\n{e}'
		# 	warnings.warn('Advertencia de error en lectura de datos cluster.') 
		# 	self.errors_log['clusters'].append(error_message)
		# # Encontramos los nombres de los vecions
		# vecinos = [d['nombre'] for d in data_vecinos]
		# Creamos el cluster
		cluster = Cluster(
			nombre=nombre,
			pozos=pozos,
			coordenadas=coordenadas,
			# vecinos=vecinos,
		)
		return cluster	

	# def agregar_rutas(
	# 			self, 
	# 			nombre_cluster:str,
	# 			data_vecinos:Dict[str, any]
	# 		) -> None:
	# 	"""
	# 	Agrega la informacion de los vecinos a los
	# 	datos de las rutas.
	# 	"""
	# 	for vecino in data_vecinos:
	# 		# Incluimos una dirección
	# 		nombre = f'{nombre_cluster} <-> {vecino["nombre"]}'
	# 		if nombre in self.datos_rutas.keys():
	# 			assert(np.isclose(self.datos_rutas[nombre]['longitud'], vecino["dist"])), f'distancias inconsistentes {nombre}'
	# 			assert(np.isclose(self.datos_rutas[nombre]['velocidad_promedio'], vecino["dist"] / vecino["tiempo"])), f'velocidades inconsistentes {nombre}'
	# 		assert(vecino["tiempo"] > 0), f'tiempo a vecino {vecino["nombre"]} es 0'
	# 		self.datos_rutas[nombre] = {
	# 			'longitud':vecino["dist"],
	# 			'velocidad_promedio':vecino["dist"] / vecino["tiempo"] * ESCALA_VELOCIDAD_CAMION
	# 		}


class CampoUtils :

	def __init__(self) -> None:
		self.cluster_de_pozo = dict()
		self.pozos_en_cluster = dict()
		self.debug = False

	def cargar_campo_desde_archivo(
				self,
				archivo_path: Union[str, Path],
				filtro_clusters:Optional[Union[None, List[str]]]=None,
				filtro_pozos:Optional[Union[None, List[str]]]=None
			) -> CampoPozos:
		# Carga el campo desde archivo
		with open(archivo_path, 'rb') as f:
			campo = pickle.load(f)
		# Genera diccionarios
		for cluster in campo.clusters:
			for pozo in cluster.pozos:
				self.cluster_de_pozo[pozo.nombre] = cluster.nombre
			self.pozos_en_cluster[cluster.nombre] = [pozo.nombre for pozo in cluster.pozos]
		if self.debug:
			print('')
			print('Pozos en clusters:')
			pozos = sorted(list(self.cluster_de_pozo.keys()))
			for pozo in pozos:
				print('\t', pozo)
		# Aplicar filtros
		if filtro_clusters is not None or filtro_pozos is not None:
			campo = self.filtrar(
				campo,
				filtro_clusters=filtro_clusters,
				filtro_pozos=filtro_pozos
			)
		return campo		

	def cargar_campo(
			self,
			campo,
			filtro_clusters:Optional[Union[None, List[str]]]=None,
			filtro_pozos:Optional[Union[None, List[str]]]=None
	) -> CampoPozos:
		# Genera diccionarios
		for cluster in campo.clusters:
			for pozo in cluster.pozos:
				self.cluster_de_pozo[pozo.nombre] = cluster.nombre
			self.pozos_en_cluster[cluster.nombre] = [pozo.nombre for pozo in cluster.pozos]
		if self.debug:
			print('')
			print('Pozos en clusters:')
			pozos = sorted(list(self.cluster_de_pozo.keys()))
			for pozo in pozos:
				print('\t', pozo)
		# Aplicar filtros
		if filtro_clusters is not None or filtro_pozos is not None:
			campo = self.filtrar(
				campo,
				filtro_clusters=filtro_clusters,
				filtro_pozos=filtro_pozos
			)
		return campo

	def filtrar(
				self, 
				campo:CampoPozos,
				filtro_clusters:Union[None, List[str]]=None,
				filtro_pozos:Union[None, List[str]]=None
			) -> CampoPozos:
		#-----------------------------
		# Filtrado de pozos
		#-----------------------------
		if filtro_clusters is None:
			filtro_clusters = list()
		if filtro_pozos is not None:
			# Encuentra clusters de los pozos dados
			filtro_pozos = self.auditar_filtro(filtro_pozos)
			pozos = sorted(list(self.cluster_de_pozo.keys()))
			pozos_no_conocidos = [pozo for pozo in filtro_pozos if pozo not in pozos]
			if len(pozos_no_conocidos) > 0:
				print('Atención: los siguientes pozos no están en el campo')
				for pozo in pozos_no_conocidos:
					print('\t', pozo)
				filtro_pozos = [pozo for pozo in filtro_pozos if pozo not in pozos_no_conocidos]
			assert(len(filtro_pozos) > 0)
			if self.debug:
				print('')
				print('Pozos en filtro:')
				for pozo in filtro_pozos:
					print('\t', pozo)
			clusters = [self.cluster_de_pozo[pozo] for pozo in filtro_pozos]
			filtro_clusters += clusters
		#-----------------------------
		# Filtra clusters
		#-----------------------------
		filtro_clusters = self.auditar_filtro(filtro_clusters)
		assert(len(filtro_clusters) > 0)
		if 'bodega_00' not in filtro_clusters: 
			filtro_clusters.insert(0, 'bodega_00')
		clusters = [cluster for cluster in campo.clusters if cluster.nombre in filtro_clusters]
		# Filtra pozos en clusters
		clusters_sin_pozo = list()
		if filtro_pozos is not None:
			for cluster in clusters:
				pozos = [pozo for pozo in cluster.pozos if pozo.nombre in filtro_pozos]
				if len(pozos) == 0 and cluster.nombre != 'bodega_00':
					clusters_sin_pozo.append(cluster.nombre)
				else:
					cluster.pozos = pozos
		clusters = [cluster for cluster in clusters if cluster.nombre not in clusters_sin_pozo]
		#-----------------------------
		# Actualiza registros
		#-----------------------------
		campo.clusters = clusters		
		campo.backup_clusters = deepcopy(clusters)
		campo.num_clusters = len(campo.clusters)
		campo.punto_partida = clusters[0].nombre
		campo.dict_clusters = {cluster.nombre:p for p, cluster in enumerate(campo.clusters)}
		campo.coordenadas = {cluster.nombre:cluster.coordenadas for cluster in campo.clusters}
		campo.actualizar_camiones()
		campo.backup_camiones = deepcopy(campo.camiones)
		campo.estado = campo.reset()
		campo.nS = campo.estado.shape[0]
		campo.comienzo = True # El entorno recién se inicializa
		campo.nA = campo.num_clusters * campo.num_camiones
		campo.debug = False
		campo.render_mode = 'rgb_array'
		return campo
	
	def auditar_filtro(
				self,
				filtro_clusters:List[str]
			) -> List[str]:
		clusters_auditados = list()
		for cluster in filtro_clusters:
			if ':1' in cluster:
				cluster = estandarizar_nombre(cluster)
			clusters_auditados.append(cluster)
		clusters_auditados = list(set(clusters_auditados))
		return sorted(clusters_auditados)


'''
© Copyright ArdillaByte Inc. 2023

-----------------------------------------------
Clase para crear el espacio de estados de 
búsqueda de rutas.
-----------------------------------------------
'''
from pathlib import Path
import pandas as pd
from typing import List, Dict, Optional, Tuple, Union
from itertools import product
from collections import defaultdict
import json
from tqdm.auto import tqdm

from chaparral.mapa.espacio_rutas import RutasClusters
from chaparral.modelos.constantes import ESCALA_TIEMPO_RUTA


class CreadorEspacioRutas :
	"""Clase para crear un espacio de estados para las rutas"""
	def __init__(
				self,
				nombre_archivo_clusters:str
			) -> None:
		# -------------------------------
		# Registros de directorios
		# -------------------------------
		self.directorio = Path.cwd() / Path('..').resolve() / Path('data', 'objetos_mapa')
		self.directorio.mkdir(parents=False, exist_ok=True)
		# -------------------------------
		# Registros de archivos
		# -------------------------------
		self.nombre_archivo_clusters = nombre_archivo_clusters
		# -------------------------------
		# Registro de datos
		# -------------------------------
		self.data = None
		self.verbose = False
	
	def crear(self) -> RutasClusters:
		# Cargar datos desde archivo
		self.cargar_datos()
		# Crea las rutas
		rutas = dict()
		# Crea las coordenadas
		coordenadas = dict()
		# Crea los tiempo_desplazamiento
		tiempo_desplazamiento = dict()
		# Iteramos sobre los registos
		for index, linea in self.data.iterrows():
			rutas[linea.nombre] = self.obtener_vecinos(linea)
			coordenadas[linea.nombre] = linea.coords
			tiempo_desplazamiento[linea.nombre] = self.obtener_tiempos(linea)
		espacio = RutasClusters(
			cluster_inicial='',
			cluster_objetivo='',
			rutas=rutas,
			coordenadas=coordenadas,
			tiempo_desplazamiento=tiempo_desplazamiento
		)
		return espacio 
	
	def cargar_datos(self) -> None:
		ruta_archivo = self.directorio / Path(self.nombre_archivo_clusters)
		self.data = pd.read_json(ruta_archivo, lines=True)
		if self.verbose:
			print('')
			print(f'Datos cargados con éxito desde archivo.\n\t({ruta_archivo})')
	
	def obtener_vecinos(self, datos:pd.Series) -> List[str]:
		vecinos = list()
		vecinos_ = datos.vecinos
		for vecino in vecinos_:
			vecinos.append(vecino['nombre'])
		return vecinos
	
	def obtener_tiempos(self, datos:pd.Series) -> List[str]:
		tiempo_desplazamiento = dict()
		vecinos = datos.vecinos
		for vecino in vecinos:
			tiempo_desplazamiento[vecino['nombre']] = vecino['tiempo'] * ESCALA_TIEMPO_RUTA
		return tiempo_desplazamiento


class CreadorBaseRutas:
	"""
	Clase para crear la base de datos de las rutas
	de cluster a cluster
	"""
	def __init__(
				self,
				nombre_archivo:str
			) -> None:
		# -------------------------------
		# Registros de directorios
		# -------------------------------
		self.directorio = Path.cwd() / Path('..').resolve() / Path('data', 'objetos_mapa')
		self.directorio.mkdir(parents=False, exist_ok=True)
		self.directorio_log = Path.cwd() / Path('..').resolve() / Path('data', 'logs')
		self.directorio_log.mkdir(parents=False, exist_ok=True)
		# -------------------------------
		# Registros de archivos
		# -------------------------------
		self.nombre_archivo = nombre_archivo
		# -------------------------------
		# Registros de datos
		# -------------------------------
		self.base_rutas = defaultdict(dict)
		self.verbose = True
		self.errors_log = list()

	def crear_base_rutas(self):
		creador = CreadorEspacioRutas(
			nombre_archivo_clusters='objetos_api.ndjson'
		)
		# Creamos el espacio de estados
		espacio = creador.crear()
		# Creamos las parejas de clusters
		clusters = list(espacio.rutas.keys())
		parejas = product(clusters, repeat=2)
		parejas = [pareja for pareja in parejas if pareja[0] != pareja[1]]
		# Iteramos sobre las parejas para encontrar
		# la ruta más corta
		for pareja in tqdm(parejas):
			cluster_desde, cluster_a = pareja
			espacio.cluster_inicial = cluster_desde
			espacio.cluster_objetivo = cluster_a
			ruta = espacio.A_star(W=1)
			if ruta is not None:
				self.base_rutas[cluster_desde][cluster_a] = ruta.costo_camino
			else:
				mensaje = f'Clusters {cluster_desde} y {cluster_a} no están conectados.'
				self.errors_log.append(mensaje)
		# Guardamos base a archivo
		self.guardar()
		# Guardamos log de errores
		json_object = json.dumps(self.errors_log, indent=4)
		nombre_archivo_log = self.directorio_log / Path('rutas_log.json')
		with open(nombre_archivo_log, "w") as outfile:
			outfile.write(json_object)
		if self.verbose:
			print(f'Log de errores guardado en:\n\t{nombre_archivo_log}')

	def guardar(
				self, 
				file:Optional[Union[None, str]]=None
			) -> None:
		nombre_archivo = self.directorio / Path('rutas.json')
		with open(nombre_archivo, "w") as outfile:
			json.dump(self.base_rutas, outfile)
		if self.verbose:
			print(f'Base de datos de rutas guardadas en:\n\t{nombre_archivo}')